#include "generic.h"
#include "vektorr2.h"
#include "vektorr3.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>
#include "SDL/SDL_gfxPrimitives.h"

using namespace std;

struct polygonEdge
{
  VektorR2 p; // point in space
  VektorR2 n; // normal vector
}; // struct polygonEdge

const int NO_CURVES  =  6;
const int NO_KNOTS   =  4;
const int SURF_RES_X = 60;

double RESIZE_FACTOR = 1.5;
bool   display_surface = false;

VektorR3    Knot[NO_CURVES][NO_KNOTS];
polygonEdge polygon_edge[NO_CURVES];

int      window_width, window_height;

void RenderSurface(SDL_Surface*, double, double);

double sqr(double v)
{
  return v*v;
} // sqr

inline int save_index(int i)
{
  if(i < 0) return NO_CURVES+i;
  return i % NO_CURVES;
} // save_index

inline void SetPixelColor(SDL_Surface *screen, int x, int y, unsigned char red, unsigned char green, unsigned char blue)
{
  if(x < 0) return;
  if(y < 0) return;

  if(x >= window_width) return;
  if(y >= window_height) return;

  ((unsigned char*)(screen->pixels))[x*3+0 + y*screen->pitch] = red;
  ((unsigned char*)(screen->pixels))[x*3+1 + y*screen->pitch] = green;
  ((unsigned char*)(screen->pixels))[x*3+2 + y*screen->pitch] = blue;
} // SetPixelColor

void SetFatPixelColor(SDL_Surface *screen, int x, int y, unsigned char red, unsigned char green, unsigned char blue)
{
  SetPixelColor(screen, x, y, red, green, blue);
  SetPixelColor(screen, x+1, y, red, green, blue);
  SetPixelColor(screen, x, y+1, red, green, blue);
  SetPixelColor(screen, x+1, y+1, red, green, blue);
} // SetFatPixelColor

inline void SetPixel(SDL_Surface* screen, int x, int y)
{
  SetPixelColor(screen, x, y, 255, 255, 255);
} // SetPixel

VektorR3 CalcHermit(VektorR3* P, double t)
{
  return (1 - 3*t*t + 2*t*t*t)*P[0] + (3*t*t - 2*t*t*t)*P[1] + (t - 2*t*t + t*t*t)*4.0*(P[0]-P[2]) + (-t*t + t*t*t)*4.0*(P[1]-P[3]);
} // CalcHermit

VektorR3 Rotate(VektorR3 point, double alpha, double beta)
{
  VektorR3 P1, P2;

  P1[0] = point[0];
  P1[1] = point[1]*cos(alpha) - point[2]*sin(alpha);
  P1[2] = point[1]*sin(alpha) + point[2]*cos(alpha);

  P2[0] = P1[0]*cos(beta) - P1[2]*sin(beta);
  P2[1] = P1[1];
  P2[2] = P1[0]*sin(beta) + P1[2]*cos(beta);

  return P2;
} // Rotate

void RenderHermit(SDL_Surface *screen, VektorR3* P, double alpha, double beta)
{
  VektorR3 point, display_point;

  for(double t = 0.0; t <= 1.0; t += 1.0/SURF_RES_X) {
    
    point = CalcHermit(P, t);
    display_point = Rotate(point*RESIZE_FACTOR, alpha, beta);
    
    SetPixelColor(screen, (int)display_point[0]+window_width/2, (int)display_point[1]+window_height/2, 255, 255, 255);     
  } // for 
} // RenderHermit

void DisplayKnots(SDL_Surface *screen, double alpha, double beta, long color)
{
  VektorR3 point;

  for(int i = 0; i <= NO_CURVES; i++) {
    
    if(i > 0) {
      aalineColor(screen, polygon_edge[i-1].p[0]*60+1, polygon_edge[i-1].p[1]*60+1, 
		  polygon_edge[i % NO_CURVES].p[0]*60+1, polygon_edge[i % NO_CURVES].p[1]*60+1, 0xFFFFFFFF);
    } // if
    
  } // for
  
  int last_knot = 6;

  for(int j = 0; j < last_knot; j++) {
    for(int i = 0; i < NO_KNOTS; i++) {

      point = Rotate(Knot[j][i]*RESIZE_FACTOR, alpha, beta);

      aalineColor(screen, (int)(point[0])-2+window_width/2, (int)(point[1])-2+window_height/2, 
		  (int)(point[0])+2+window_width/2, (int)(point[1])+2+window_height/2, color);
      aalineColor(screen, (int)(point[0])+2+window_width/2, (int)(point[1])-2+window_height/2, 
		  (int)(point[0])-2+window_width/2, (int)(point[1])+2+window_height/2, color);
    } // for
  } // for
  
  for(int i = 0; i < last_knot; i++) {
    RenderHermit(screen, Knot[i], alpha, beta);
  } // for

  if(display_surface)
    RenderSurface(screen, alpha, beta);

} // DisplayKnots

void IndexClosestPoint(int x, int y, double alpha, double beta, int &index_i, int &index_j)
{
  double   dist, min_dist = 10000000.0;
  int      index;
  VektorR3 cursor = VektorR3(x-window_width/2, y-window_height/2, 0), point;

  for(int j = 0; j < NO_CURVES; j++) {
    for(int i = 0; i < NO_KNOTS; i++) {

      point = Rotate(Knot[j][i]*RESIZE_FACTOR, alpha, beta);

      dist = (sqr(cursor[0] - point[0])+sqr(cursor[1] - point[1]));
    
      if(min_dist > dist) {
	index_i = i;
	index_j = j;
	min_dist = dist;
      } // if
    } // for
  } // for

} // IndexClosestPoint

double CalcArea(VektorR2 A, VektorR2 B, VektorR2 C)
{
  double a = (C-B).Norm2();
  double b = (A-C).Norm2();
  double c = (A-B).Norm2();

  double s = (a+b+c)/2.0;

  return sqrt(s*(s-a)*(s-b)*(s-c));
} // CalcArea

bool DetermineBarrycentricCoords(double x, double y, double l[])
{
  VektorR2 direction, point;
  double   alpha[NO_CURVES], pi[NO_CURVES];
  bool     outside;

  // is this point inside the polygon?
  
  point = VektorR2(x, y);

  for(int i = 0; i < NO_CURVES; i++) {
    if((polygon_edge[i].p - point) * polygon_edge[i].n > 0.0)
      return false;
  } // for

  // yes, (x, y) is inside the domain polygon

  // => insert your code now ... (you may want to use the function
  //    CalcArea(point_1, point_2, point_3))

  //    The barrycentric coordinates to return should be stored in l[]

  return true;
} // DetermineBarrycentricCoords

bool DetermineSurfacePoint(double x, double y, VektorR3 &Q)
{
  VektorR2 direction, point;
  double   l[NO_CURVES];
  bool     outside;

  if(DetermineBarrycentricCoords(x, y, l) == true) {

    // caluclate point Q on patch

    // => insert your code here

    return true;
  } // if

  return false;
} // DetermineSurfacePoint

void RenderSurface(SDL_Surface *screen, double rot_a, double rot_b)
{
  for(double x = 0.0; x <= 2.0; x += 0.04) {
    for(double y = 0.0; y <= 2.0; y += 0.04) {

      VektorR3 s_point;

      if(DetermineSurfacePoint(x, y, s_point) == true) {

	VektorR3 d_point = Rotate(s_point*RESIZE_FACTOR, rot_a, rot_b);

	SetFatPixelColor(screen, (int)(d_point[0])+window_width/2, (int)(d_point[1])+window_height/2, 
			 (4-x-y)*255/4.0, y*255/2.0, x*255/2.0);
	SetPixelColor(screen, x*60+1, y*60+1, (4-x-y)*255/4.0, y*255/2.0, x*255/2.0);
      } // if
    } // for
  } // for
} // RenderSurface

int main(int argc, char** argv)
{
    SDL_Surface *screen, *backup_screen;
    SDL_Event   event;
    SDL_Rect    dst_rect, full_image_rect, scrolled_rect, location;
    int         thumb_height, thumb_border_size;
    long        pixel_set = 0, index;
    bool        l_button_down = false, r_button_down = false;
    double      alpha, beta;
    VektorR3    knot, cursor;
    int         index_i, index_j;
    int         cursor_x, cursor_y;
    VektorR3    s_point, d_point;

    window_width = 550; window_height = 550;

    Knot[0][0] = VektorR3(-100, 0 , 75);
    Knot[0][1] = VektorR3(100, 0 , 75);
    Knot[0][2] = VektorR3(-120, -40 , 75);
    Knot[0][3] = VektorR3(110, 40 , 55);
    
    Knot[1][0] = VektorR3(100, 0 , 75);
    Knot[1][1] = VektorR3(145, 0 , 0);
    Knot[1][2] = VektorR3(80, -5 , 60);
    Knot[1][3] = VektorR3(150, -65 , -5);
    
    Knot[2][0] = VektorR3(145, 0 , 0);
    Knot[2][1] = VektorR3(100, 0 , -75);
    Knot[2][2] = VektorR3(125, -15 , 15);
    Knot[2][3] = VektorR3(80, -30 , -85);
    
    Knot[3][0] = VektorR3(100, 0 , -75);
    Knot[3][1] = VektorR3(-80, 0 , -65);
    Knot[3][2] = VektorR3(115, -50 , -70);
    Knot[3][3] = VektorR3(-85, 50 , -85);
    
    Knot[4][0] = VektorR3(-80, 0 , -65);
    Knot[4][1] = VektorR3(-110, 0 , 0);
    Knot[4][2] = VektorR3(-30, -35 , -85);
    Knot[4][3] = VektorR3(-70, 5 , 5);
    
    Knot[5][0] = VektorR3(-110, 0 , 0);
    Knot[5][1] = VektorR3(-100, 0 , 75);
    Knot[5][2] = VektorR3(-80, -10 , -10);
    Knot[5][3] = VektorR3(-130, -25 , 15);

    if(SDL_Init(SDL_INIT_VIDEO) < 0) {
	cerr << "Could not initialize SDL video" << endl;
	exit(-1);
    } // if
    atexit(SDL_Quit);

    Uint32 flags = 0;

    // flags |= SDL_FULLSCREEN;

    screen = SDL_SetVideoMode(window_width, window_height, 24, flags);

    if(!screen) {
	cerr << "could not SetVideoMode() - abort" << endl;
	exit(-1);
    } // if

    VektorR2 direction;
    
    // generate patch
    
    for(int i = 0; i <= NO_CURVES; i++) {
      
      if(i < NO_CURVES)
	polygon_edge[i].p = VektorR2(sin(((double)i*2*M_PI/(double)NO_CURVES))+1.0,
				     cos(((double)i*2*M_PI/(double)NO_CURVES))+1.0);
      
      if(i > 0) {
	aalineColor(screen, polygon_edge[i-1].p[0]*60+1, polygon_edge[i-1].p[1]*60+1, 
		    polygon_edge[i % NO_CURVES].p[0]*60+1, polygon_edge[i % NO_CURVES].p[1]*60+1, 0xFFFFFFFF);
	
	direction = polygon_edge[i % NO_CURVES].p - polygon_edge[i-1].p;
	
	polygon_edge[i-1].n = VektorR2(direction[1], -1.0*direction[0]);
	
	if(polygon_edge[i-1].n * (polygon_edge[i-1].p-VektorR2(1.0, 1.0)) > 0.0) {
	  polygon_edge[i-1].n *= -1.0;
	} // if
	
      } // if
      
    } // for

    DisplayKnots(screen, alpha, beta, 0xFFFFFF);
   
    int min_x, max_x, min_y, max_y;
    int old_min_x, old_max_x, old_min_y, old_max_y;

    old_min_x = 4; old_max_x = window_width-4; old_min_y = 4; old_max_y = window_height-4;

    for(;;) {
        
      if((SDL_WaitEvent(&event) == 0) && (event.type == SDL_KEYDOWN)) 
	goto exit_app;

      switch(event.type) {
	
      case SDL_MOUSEMOTION:	      
	cursor_x = event.motion.x;
	cursor_y = event.motion.y;

	if(l_button_down) {
	  
	  bzero(screen->pixels, screen->pitch*window_height);
	  DisplayKnots(screen, alpha, beta, 0xFFFFFFFF);

	  if((cursor_x <= 120) && (cursor_y <= 120)) {
	    SetFatPixelColor(screen, cursor_x, cursor_y, 255, 255, 255);

	    if(DetermineSurfacePoint(cursor_x*2.0/120.0, cursor_y*2.0/120.0, s_point) == true) {
	      
	      d_point = Rotate(s_point*RESIZE_FACTOR, alpha, beta);

	      aacircleColor(screen, (int)(d_point[0])+window_width/2, (int)(d_point[1])+window_height/2, 10, 0xFFFFFFFF);
	    } // if
	  } // if
	  else {
	    beta = (double)(event.motion.x-window_width/2)*3.19/(double)window_width;
	    alpha = (double)(event.motion.y-window_height/2)*3.19/(double)window_height;	    
	  } // if

	  SDL_UnlockSurface(screen);
	  SDL_UpdateRect(screen, 0, 0, window_width-1, window_height-1);
	  SDL_LockSurface(screen);
	} // if

	while(SDL_PollEvent(&event)) {
	} // while

	break;
	
      case SDL_MOUSEBUTTONDOWN:
	l_button_down = true;
	break;
	
      case SDL_MOUSEBUTTONUP:
	l_button_down = false;
	break;
	
      case SDL_KEYDOWN:
	
	switch(event.key.keysym.sym) {

	case 'p':
	  for(int j = 0; j < NO_CURVES; j++) {
	    for(int i = 0; i < NO_KNOTS; i++) {

	      cout << "Knot[" << j << "][" << i << "] = VektorR3(" << Knot[j][i][0] << ", " << Knot[j][i][1] << " , " << Knot[j][i][2] << ");" << endl;
	    } // for

	    cout << endl;
	  } // for
	  break;
	  
	case 's':
	  display_surface = !display_surface;
	  break;
	  
	case '1':
	  bzero(screen->pixels, screen->pitch*window_height);
	  IndexClosestPoint(cursor_x, cursor_y, alpha, beta, index_i, index_j);
	  
	  knot = Knot[index_j][index_i];
	  //	  cursor = Rotate(VektorR3(event.motion.x-256, event.motion.y-256, 0.0), alpha, beta);

	  knot[0] += 10;
	  Knot[index_j][index_i] = knot;
	  
	  DisplayKnots(screen, alpha, beta, 0xFFFFFFFF);

	  SDL_UnlockSurface(screen);
	  SDL_UpdateRect(screen, 0, 0, window_width-1, window_height-1);
	  SDL_LockSurface(screen);
	  break;

	case '2':
	  bzero(screen->pixels, screen->pitch*window_height);
	  IndexClosestPoint(cursor_x, cursor_y, alpha, beta, index_i, index_j);
	  
	  knot = Knot[index_j][index_i];
	  //	  cursor = Rotate(VektorR3(event.motion.x-256, event.motion.y-256, 0.0), alpha, beta);

	  knot[0] -= 10;
	  Knot[index_j][index_i] = knot;
	  
	  DisplayKnots(screen, alpha, beta, 0xFFFFFFFF);

	  SDL_UnlockSurface(screen);
	  SDL_UpdateRect(screen, 0, 0, window_width-1, window_height-1);
	  SDL_LockSurface(screen);
	  break;

	case '3':
	  bzero(screen->pixels, screen->pitch*window_height);
	  IndexClosestPoint(cursor_x, cursor_y, alpha, beta, index_i, index_j);
	  
	  knot = Knot[index_j][index_i];
	  //	  cursor = Rotate(VektorR3(event.motion.x-256, event.motion.y-256, 0.0), alpha, beta);

	  knot[1] += 10;
	  Knot[index_j][index_i] = knot;
	  
	  DisplayKnots(screen, alpha, beta, 0xFFFFFFFF);

	  SDL_UnlockSurface(screen);
	  SDL_UpdateRect(screen, 0, 0, window_width-1, window_height-1);
	  SDL_LockSurface(screen);
	  break;

	case '4':
	  bzero(screen->pixels, screen->pitch*window_height);
	  IndexClosestPoint(cursor_x, cursor_y, alpha, beta, index_i, index_j);
	  
	  knot = Knot[index_j][index_i];
	  //	  cursor = Rotate(VektorR3(event.motion.x-256, event.motion.y-256, 0.0), alpha, beta);

	  knot[1] -= 10;
	  Knot[index_j][index_i] = knot;
	  
	  DisplayKnots(screen, alpha, beta, 0xFFFFFFFF);

	  SDL_UnlockSurface(screen);
	  SDL_UpdateRect(screen, 0, 0, window_width-1, window_height-1);
	  SDL_LockSurface(screen);
	  break;

	case 27:
	case 'q':
	case 'Q':
	  goto exit_app;
	  
	default:
	  break;
	} // switch
      } // switch

      //      while(SDL_PollEvent(&event) != 0)
      //	SDL_WaitEvent(&event);
      
    } // for
	    
 exit_app:

    SDL_UnlockSurface(screen);	    
    SDL_FreeSurface(screen);

    exit(0);
} // main
